
class Set_Null:
    def __init__(self):
        self.sql = "SET NULL"


class Set_Default:
    def __init__(self):
        self.sql = "SET DEFAULT"


class Cascade:
    def __init__(self):
        self.sql = "CASCADE"


def type_to_sql_type(type):
    if type == float:
        sql_type = "REAL"
    elif type == int:
        sql_type = "INTEGER"
    elif type == str:
        sql_type = "STRING"
    elif type == bool:
        sql_type = "BOOLEAN"
    return sql_type


def sql_args(column):
    args_str = ""
    if column.unique:
        args_str += " UNIQUE"
    if column.not_null:
        args_str += " NOT NULL"
    if column.default:
        if type(column.default) == str:
            args_str += f" DEFAULT [{column.default}]"
        else:
            args_str += f" DEFAULT {column.default}"
    return args_str


def lenght_sql(column):
    lenght_str = ""
    if column.max_length and not column.min_length:
        lenght_str += f" ({column.max_length})"
    if column.min_length and column.min_length:
        lenght_str += f"({column.min_length}, {column.max_length})"
    return lenght_str


class BaseColumn:
    def __init__(self, pk=False, foreign_model=None, on_delete=None, unique=False, not_null=False, default=None, max_length=None, min_length=None):
        self.name = None
        self.pk = pk
        self.fk = True if foreign_model else False
        self.foreign_model = foreign_model
        self.on_delete = on_delete
        self.unique = unique
        self.not_null = not_null
        self.default = default
        self.max_length = max_length
        self.min_length = min_length
        self.type = type_to_sql_type(self.python_type) + lenght_sql(self)

    def generate_sql(self):
        return f"{self.name} {self.type}{sql_args(self)}"


class PrimaryKeyColumn(BaseColumn):
    def __init__(self, autoincrement=False, type=int):
        self.python_type = type
        self.autoincrement = autoincrement
        super().__init__(pk=True)

    def generate_sql(self):
        return f"{self.name} {self.type} PRIMARY KEY{' AUTOINCREMENT' if self.autoincrement else ''}"


class ForeignKeyColumn(BaseColumn):
    def __init__(self, foreign_model, on_delete=Cascade(), type=int, unique=False, not_null=False, default=None, max_length=None, min_length=None):
        self.python_type = type
        super().__init__(foreign_model=foreign_model, on_delete=on_delete, unique=unique, not_null=not_null, default=default, max_length=max_length, min_length=min_length)

    def generate_sql(self):
        self.foreign_model.get_column(self.foreign_model)
        return f"{self.name} {self.type}{sql_args(self)} REFERENCES {self.foreign_model.model.name} ({self.foreign_model.primary_key(self.foreign_model)})  ON DELETE {self.on_delete.sql}"


class IntegerColumn(BaseColumn):
    def __init__(self, unique=False, not_null=False, default=None, max_length=None, min_length=None):
        self.python_type = int
        super().__init__(unique=unique,  not_null=not_null, default=default, max_length=max_length, min_length=min_length)


class StringColumn(BaseColumn):
    def __init__(self, unique=False, not_null=False, default=None, max_length=None, min_length=None):
        self.python_type = str
        super().__init__(unique=unique,  not_null=not_null, default=default, max_length=max_length, min_length=min_length)


class FloatColumn(BaseColumn):
    def __init__(self, unique=False, not_null=False, default=None, max_length=None, min_length=None):
        self.python_type = float
        super().__init__(unique=unique,  not_null=not_null, default=default, max_length=max_length, min_length=min_length)


class BooleanColumn(BaseColumn):
    def __init__(self, unique=False, not_null=False, default=None, max_length=None, min_length=None):
        self.python_type = bool
        super().__init__(unique=unique,  not_null=not_null, default=default, max_length=max_length, min_length=min_length)



from .model import Model, ModelFunc
import time
import os
import platform
import datetime
import imp
import sqlite3

version = "0.2.9.1"
testing = False
beta = False
if testing:
    version += "-test"
elif beta:
    version += "-beta"
version_str = f"SQL-Cable Build {version}"
package_url = "url"


class Session:
    def __init__(self, db_path):
        self.db_path = db_path
        self.conn = sqlite3.connect(self.db_path)
        self.c = self.conn.cursor()

    def add(self, obj):
        obj.add_query(self.c)

    def save(self):
        self.conn.commit()

    def select(self, selectobj):
        return selectobj.execute(self.c)

    def stop(self):
        self.conn.close()


class DB:
    def __init__(self, app=None):
        self.models = []
        if platform.system() == 'Windows':
            self.path_break = '\\'
        else:
            self.path_break = '/'
        if app:
            self.init(app)
        if testing or beta:
            print(f"""WARNING you are running a {'test' if testing else 'beta'} build of SQL-Cable build version: {version}
This build maybe unstable and have bugs or other problems
Please update to the latest full build of SQL-Cable by
typing 'pip install sql-cable' in the cmd""")

    def init(self, app):
        self.app = app
        self.db_path = app.config['DB_PATH']
        self.session = Session(self.db_path)

    def register_model(self, model):
        model.db_path = self.db_path
        model.session = self.session
        model.check_for_changes(model, self.session)
        # print(str(model).split('.')[1].replace("'>", ''))
        model.load_querrys(model)
        self.models.append(model)

    def make_migrations(self, path=None):
        mirgrations = {}
        for model in self.models:
            if model.check_for_changes(model, self.session) != []:
                mirgrations[model.model.name] = model.generate_migrations(model)
        if mirgrations != {}:
            if not os.path.isdir('migrations'):
                os.mkdir('migrations')
                with open(path if path else f"migrations{self.path_break}__init__.py", 'w') as f:
                    pass
            all_files = [f for f in os.listdir('migrations') if os.path.isfile(os.path.join('migrations', f))]
            current_number = 0
            for file in all_files:
                if str(file).split('_')[0] == 'migration':
                    if int(str(file).split('_')[1].split('.')[0]) > current_number:
                        current_number = int(str(file).split('_')[1].split('.')[0])
            file_str = f"# Generated by {version_str} on {datetime.datetime.now()}\nfrom sql_cable.migrations import Migration\n"

            model_names = []

            for migration in mirgrations:
                if mirgrations[migration]:
                    model_name = migration
                    model_names.append(model_name)
                    file_str += mirgrations[migration]

            for model_name in model_names:
                file_str += f"\n\n\nmodels = {model_names}"
            with open(path if path else f"migrations{self.path_break}migration_{current_number + 1}.py", 'w') as f:
                f.write(file_str)
            file_name = f"migration_{current_number + 1}"
            with open(path if path else f"migrations{self.path_break}__init__.py", 'a') as f:
                f.write(f"\nfrom . import {file_name}")
        else:
            print("No migrations to make.")

    def migrate(self, path=None):
        if not os.path.isdir('migrations'):
            print("No dir named migrations could be found.")
        else:
            all_files = [f for f in os.listdir('migrations') if os.path.isfile(os.path.join('migrations', f))]
            current_number = 0
            for file in all_files:
                if str(file).split('_')[0] == 'migration':
                    if int(str(file).split('_')[1].split('.')[0]) > current_number:
                        current_number = int(str(file).split('_')[1].split('.')[0])
            file_name = f"migration_{current_number}"
            migrate_file = __import__(f"migrations")
            _locals = locals()
            exec(f"models = migrate_file.{file_name}.models", globals(), _locals)
            for model in _locals['models']:
                exec(f"model_in = migrate_file.{file_name}.{model}", globals(), _locals)
                _locals['model_in']().run_migration(self.db_path)
